#!/usr/bin/perl
use IO::Handle;


# read the config file
do "config.pl" || die "Can't read config.pl!\n";

#set up ...
$|=1;
srand(time|$$);
$stopme=0;
$backupnow=0;

#set up pid
open(FILE,">sys3.pid") or die "can't open pid file!";
print FILE $$;
close(FILE);

#interupt handler
sub handler {
	print "\nBANG (INT)!\n";
	$stopme=1;
}
sub backuphandler {
	print "\nBANG! (HUP)\n";
	$backupnow=1;
}

#get core (family) name out of a warrior without lcoation
sub getname {
	foreach $n (split(/\n/,$_[0])) {
		if($n=~/^;name ([^ ]+) /) {
			return $1;
		}
	}
}

#get size of warrior
sub getsize {
	my $size=0;
	foreach $n (split(/\n/,$_[0])) {
		unless($n=~/;buffer/) {
			$size++ if($n=~/^L[0-9]/);
		}
	}
	return $size;
}

#make new warrior out of an existing one
sub makenew {
	my $w="";
	my $line=0;
	my $size=getsize($_[0]);
	foreach $n (split(/\n/,$_[0])) {
		if($n=~/^;name:/) {
			$w.=";name $_[1]";
		} elsif(($n=~/^L[0-9]/) && !($n=~/;buffer/)) {
			#maybe we can do mutations by part later
			($label,$opcode,$mod,$mode1,$num1,$mode2,$num2)=
				split(/[\.,\s]+/,$n);
			$nodelete=1;
			#size change (addition/deletion)
			if(int(rand(100))<$SIZECHANGECHANCE) {
				#addition
				if(int(rand(2))==1) {
					if($size<$WARSIZE
						&& $line<($WARSIZE-1)) {
						$size++;
						$w.=MakeRandOpcode($size,
							$line) . "\n";
						$line++;
					}
				#deletion
				} else {
					if($size>1) {
						$nodelete=1;
						$size--;
					}
				}
			}
			if($nodelete && $line<$WARSIZE) {
				if(int(rand(100))<$MUTATIONCHANCE) {
					$w.=MakeRandOpcode($size,
						$line) . "\n";
				} else {
					$w.="L$line\t$opcode.$mod\t" .
						"$mode1 $num1,\t$mode2 $num2\n";
				}
				$line++;
			}
		} elsif(!($n=~/;buffer/)) {
			$w.="$n\n";
		}
	}
	for(;$line<$WARSIZE;$line++) {
		$w.="L$line\tDAT.F\t\$ 0,\t\$ 0\t;buffer\n";
	}
	return $w;
}

#get the right enviroment for this location
sub getenviro {
	my $x=shift;
	my $y=shift;
	return $NWENV if($x<$NWCUR         && $y>($SIZE-$NWCUR)); #NW
	return $NEENV if($x>($SIZE-$NECUR) && $y>($SIZE-$NECUR)); #NE
	return $SWENV if($x<$SWCUR         && $y<$SWCUR);         #SW
	return $SEENV if($x>($SIZE-$SECUR) && $y<$SECUR);         #SE
	return $DEFENV;
}

#is thsi location empty or full or not (1 for full 0 for empty),
#returns 1 if it is ... 0 if not (or out of range)
sub isit {
	my $x=shift;
	my $y=shift;
	my $what=shift;
	return 0 if($x<0 || $y<0 || $x>=$SIZE || $y>=$SIZE);
	return 1 if($what && ($warrior[$x][$y] ne ""));
	return 1 if(!($what) && ($warrior[$x][$y] eq ""));
	return 0;
}

#get a random location (1 to get a full one 0 to get an ampty one)
#returns 1,2,3,4 for N,E,S,W or 0 if no such place
sub nearother {
	my @a=();
	push(@a,1) if(isit($x,$y+1,$_[0])); #N
	push(@a,2) if(isit($x-1,$y,$_[0])); #E
	push(@a,3) if(isit($x,$y-1,$_[0])); #S
	push(@a,4) if(isit($x+1,$y,$_[0])); #W
	return 0 if($#a==0);
	return $a[int(rand($#a+1))];
}

#save warriors (done as a subroutine on exit)
sub savewar {
	print "Writing warriors ...\n";
	for($x=0;$x<$SIZE;$x++) {
		for($y=0;$y<$SIZE;$y++) {
			printf("%5d %5d\r",$x,$y);
			unlink("warriors/$x/$y.red");
			unlink("warriors/$x/$y.score");
			if($warrior[$x][$y] ne "") {
				open(FILE,">warriors/$x/$y.red")
					or warn "(can't open warrior $x,$y)";
				print FILE $warrior[$x][$y];
				close(FILE);
				open(FILE,">warriors/$x/$y.score")
					or warn "(can't open warrior " .
						"score $x,$y)";
				print FILE "Size $warriorsize[$x][$y]\n";
				print FILE "Score $warriorscore[$x][$y]\n";
				print FILE "Fights $warriorfights[$x][$y]\n";
				print FILE "Kids $warriorkids[$x][$y]\n";
				print FILE "Generation " .
					"$warriorgeneration[$x][$y]\n";
				close(FILE);
			}
		}
	}
}

print "Reading warriors ...\n";

#read in warriors
for($x=0;$x<$SIZE;$x++) {
	for($y=0;$y<$SIZE;$y++) {
		printf("%5d %5d\r",$x,$y);
		if(open(FILE,"warriors/$x/$y.red")) {
			while(<FILE>) {
				$warrior[$x][$y].=$_;
				$warriorname[$x][$y]=$1 if(/^;name (.*)$/);
			}
			close(FILE);
			open(FILE,"warriors/$x/$y.score")
				or die "(can't open warrior score $x,$y)";
			while(<FILE>) {
				if(/^Size ([0-9]+)$/) {
					$warriorsize[$x][$y]=$1;
				} elsif(/^Score ([0-9.]+)$/) {
					$warriorscore[$x][$y]=$1;
				} elsif(/^Fights ([0-9]+)$/) {
					$warriorfights[$x][$y]=$1;
				} elsif(/^Kids ([0-9]+)$/) {
					$warriorkids[$x][$y]=$1;
				} elsif(/^Generation ([0-9]+)$/) {
					$warriorgeneration[$x][$y]=$1;
				}
			}
			close(FILE);
		} else {
			$warrior[$x][$y]="";
		}
	}
}

print "Setting up interupt handler...\n";
$SIG{'INT'} = 'handler';
$SIG{'HUP'} = 'backuphandler';

##########################################################################
# main loop

print "Running ...\n";
$dobakcup=0;
do {
	#print "Finding random location ";
	do {
		$x=int(rand($SIZE));
		$y=int(rand($SIZE));
		#print ".";
	} while(($x==$y) || ($warrior[$x][$y] eq ""));
	#print "\n";

	$enviro=getenviro($x,$y);

	#fight
	if($tofight=nearother(1)) {
		$tfx=$x; $tfy=$y;
		$tfy++ if($tofight==1); #N
		$tfx-- if($tofight==2); #E
		$tfy-- if($tofight==3); #S
		$tfx++ if($tofight==4); #W
		die "warrior empty! nearother error"
			if($warrior[$tfx][$tfy] eq "");
		#print "Fight: $warriorname[$x][$y] vs. " .
			#"$warriorname[$tfx][$tfy]: ";
		pipe(IN1,OUT1) || die "no pipe";
		OUT1->autoflush(1);
		pipe(IN2,OUT2) || die "no pipe";
		OUT2->autoflush(1);
		print OUT1 $warrior[$x][$y];
		close(OUT1);
		print OUT2 $warrior[$tfx][$tfy];
		close(OUT2);
		$fdpipe1="/dev/fd/" . fileno(IN1);
		$fdpipe2="/dev/fd/" . fileno(IN2);
		$sc1=0;
		$sc2=0;
		foreach $n (split(/\n/,
			`$PMARS -b -r $FIGHTS $enviro $fdpipe1 $fdpipe2`)) {
			if($n=~/Results: ([0-9]+) ([0-9]+) ([0-9]+)/) {
				$sc1=$1*3+$3;
				$sc2=$2*3+$3;
			}
		}
		if($sc1==0 && $sc2==0) {
			print "------------\n";
			print $warrior[$x][$y];
			print "------------\n";
			print $warrior[$tfx][$tfy];
			print "------------\n";
			warn "noresults, probably pmars error";
		}
		$sc1=($sc1*100)/$FIGHTS;
		$sc2=($sc2*100)/$FIGHTS;
		close(IN1);
		close(IN2);
		if($sc1>$sc2) {
			#print "$warriorname[$x][$y] won\n";
			$warrior[$tfx][$tfy]="";
			$warriorscore[$x][$y]=
				(($warriorscore[$x][$y]*
				$warriorfights[$x][$y])+$sc1)/
				($warriorfights[$x][$y]+1);
			$warriorfights[$x][$y]++;
		} elsif($sc1<$sc2) {
			#print "$warriorname[$tfx][$tfy] won\n";
			$warrior[$x][$y]="";
			$warriorscore[$tfx][$tfy]=
				(($warriorscore[$tfx][$tfy]*
				$warriorfights[$tfx][$tfy])+$sc1)/
				($warriorfights[$tfx][$tfy]+1);
			$warriorfights[$tfx][$tfy]++;
		} else {
			#print "tie\n";
			$warriorscore[$x][$y]=
				(($warriorscore[$x][$y]*
				$warriorfights[$x][$y])+$sc1)/
				($warriorfights[$x][$y]+1);
			$warriorfights[$x][$y]++;
			$warriorscore[$tfx][$tfy]=
				(($warriorscore[$tfx][$tfy]*
				$warriorfights[$tfx][$tfy])+$sc1)/
				($warriorfights[$tfx][$tfy]+1);
			$warriorfights[$tfx][$tfy]++;
		}
	}

	#reproduce (if there is space and if this warrior still exists)
	if($warrior[$x][$y] ne "" && ($toreproduce=nearother(0))) {
		$trx=$x; $try=$y;
		$try++ if($toreproduce==1); #N
		$trx-- if($toreproduce==2); #E
		$try-- if($toreproduce==3); #S
		$trx++ if($toreproduce==4); #W
		#print "Reproducing: $warriorname[$x][$y] at $trx,$try\n";
		$warriorkids[$x][$y]++;
		$warrior[$trx][$try]=makenew($warrior[$x][$y],
			getname($warrior[$x][$y]) . " ($trx,$try)");
		$warriorname[$trx][$try]=getname($warrior[$x][$y]) .
			"($trx,$try)";
		$warriorgeneration[$trx][$try]=$warriorgeneration[$x][$y]+1;
		$warriorsize[$trx][$try]=getsize($warrior[$x][$y]);
		$warriorkids[$trx][$try]=0;
		$warriorscore[$trx][$try]=0;
		$warriorfights[$trx][$try]=0;
	}



	if($dobackup>$BACKUPFREQ || $backupnow) {
		print "Backup invoked...\n";
		savewar();
		$dobackup=0;
		$backupnow=0;
		print "\nBackup done...\n";
	}
	$dobackup++;
} until($stopme);
print "\n";
savewar();
print "Bye!\n";
